/* IluJava_IluRT0.c */
/* Chris Jacobi, April 18, 1997 3:37 pm PDT */

/*
 * Copyright (c) 1996, 1997 Xerox Corporation.  All Rights Reserved.  
 * Unlimited use, reproduction, and distribution of this software is
 * permitted.  Any copy of this software must include both the above
 * copyright notice of Xerox Corporation and this paragraph.  Any
 * distribution of this software must comply with all applicable United
 * States export control laws.  This software is made available AS IS,
 * and XEROX CORPORATION DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
 * INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE, AND NOTWITHSTANDING ANY OTHER
 * PROVISION CONTAINED HEREIN, ANY LIABILITY FOR DAMAGES RESULTING FROM
 * THE SOFTWARE OR ITS USE IS EXPRESSLY DISCLAIMED, WHETHER ARISING IN
 * CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, EVEN IF
 * XEROX CORPORATION IS ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
 */
 
/* $Id: IluJava_IluRT0.c,v 1.25 1997/04/18 23:06:50 jacobi Exp $ */

/*
 * C side of IluRT0.java
 */

#include <iluxport.h>
#include <iluerror.h>
#include <iluhash.h>
#include <iluntrnl.h>
#include <iluerrs.h>

#include "IluJava_JStubs.h"
#include "IluJava_CJ.h"
#include "IluJava.h"
#include "IluJava_JTypes.h"
#include "IluJava_Ops.h"


IMPORTVAR char * _ilujava_notUsed1;
IMPORTVAR char * _ilujava_notUsed2;

/* ================ Registry ================ */


static HashTable _c2j; /* key ilu_Class; val is java IluClassRep */


EXPORTLIMITED void
_ilujava_IluRt0InitHashTable()
{
    _c2j = ilu_hash_MakeNewTable(111, 
    	ilu_hash_HashPointer, 
    	ilu_hash_PointerCompare
    	);
}

 
/* locking by caller */
EXPORTLIMITED void 
_ilujava_registerJavaClass (
	ilu_Class cIluClass, 
	/* JIluClassRep */ ilu_refany jjClassRep
	)
/* Register java class necessary to CREATE java surrogate object or TRUE */
{
  ilu_hash_AddToTable(_c2j, 
  	(ilu_refany) cIluClass, 
  	(ilu_refany) jjClassRep
  	);
}


EXPORTLIMITED ilu_refany   /* JIluClassRep */
_ilujava_findJavaClass(ilu_Class cIluClass)
/* Get java class necessary to CREATE java surrogate object */
{
    return ilu_hash_FindInTable(_c2j, cIluClass);
}

/* ================ ======== ================ */


  
/* Inside(cIluObject->server, staticClass) */
/* oi must be finished */
EXPORTLIMITED ilu_refany /* JIluOInt */
_ilujava_creatSurrOIFromRegistry(
	ilu_Object cIluObject, 
	ilu_refany jjClassRep   /* JIluClassRep (NIL IS OK) */
	)
{
    JIluClassRep jExpectedClass 
    	= (JIluClassRep) jjClassRep;
    HObject * jobj;
    JIluOInt jjoi;
    ilu_Class cActualIluClass;
    JIluClassRep jActualClass;
    
    cActualIluClass = ilu_ClassOfObject(cIluObject);
    jActualClass = (JIluClassRep)
        _ilujava_findJavaClass(cActualIluClass);
    if (jActualClass==0) {
        SignalError(EE(), "xerox/ilu/IluSomeSystemException", 
            "surrogate class not registered"
            );
        BUGPRINT("$ IluJava failure: surrogate class not found... \n");
        return 0;
    }
/*
    if (jActualClass != jExpectedClass) {
        SignalError(EE(), "xerox/ilu/IluSomeSystemException", "ilu-class missmatch");
        BUGPRINT("$ IluJava failure: class missmatch... \n");
        return 0;
    }
*/
    jobj = execute_java_constructor(
    	PRIVILEGED_EE,
    	"xerox/ilu/IluOInt", 
    	0, /* ClassClass * or 0 if named */
    	"()"
    	);
    if (jobj==0) 
    {
       /* failed somehow in java */
       _ilujava_PANIC("$ IluJava: IluOInt creation failed... \n");
       return 0;
    }
    jjoi = (JIluOInt) jobj;
    unhand(jjoi)->jjClassRep = jActualClass;
    Access(jjoi, yServer, ilu_Server) = ilu_ServerOfObject(cIluObject);
    Access(jjoi, yIluKernelObject, ilu_Object) = cIluObject;
    ilu_RegisterLanguageSpecificObject(cIluObject, jjoi, JAVALANGIDX);
    /* We just created jjoi. Don't declare it a ghost; if it is found
     * by the collector, so be it.
     */
    return (ilu_refany) jjoi;
}

 
/* compare this to /tilde/jacobi/ilus/runtime/c/ilu.c 
 * and look at /tilde/jacobi/ilus/runtime/kernel/iluerrs.h
 */ 

EXPORTLIMITED void 
_ilujava_Report(ilu_Error* errPtr, char * t)
/* OPTIONALLY (according to command line options)
 * makes a call to ilu_DebugPrintf.
 * Does not raise any errors or exceptions.
 * Explicitely also supports errPtr==0.
 */
{
    if (_ilujava_iluGenericFlag > 0) {
        if (errPtr) {
            ilu_DebugPrintf("$ Report: %s %s %s \n", 
                t, ILU_ERR_NAME(*errPtr), ILU_ERR_DESCRIPTION(*errPtr)
                );
        } else {
            ilu_DebugPrintf("$ Report: %s \n", t);
        }
    }
}


EXPORTLIMITED void 
_ilujava_IluErrorToCallException(ilu_Error* errPtr, char* t)
/* Like IluErrorToException except that there is no switch
 * on the exception but allways IluCallException is raised.
 */
{
    if (t==0) t = "unspecified";
    _ilujava_Report(errPtr, t);
    SignalError(EE(), "xerox/ilu/IluCallException", t);
}


#define CORBAPREFIX "CORBA/corba_"

/* Shorthand used in IluErrorToException */
#define E_SWITCH(errkind, exc) 				\
    ILU_ERR_CASE(errkind, v) {				\
        SignalError(EE(), exc, t); 			\
        return;						\
    }

/* Shorthand used in IluErrorToException 
 * IN CASE THE EXCEPTION IS NOT YET IMPLEMENTED
 */
#define UNIMPL_SWITCH(errkind, ignored) 			\
    ILU_ERR_CASE(errkind, v) {				\
        SignalError(EE(), "xerox/ilu/IluSomeSystemException", t); 	\
        return;						\
    }

EXPORTLIMITED void 
_ilujava_IluErrorToException(ilu_Error* errPtr, char * t)
{
    _ilujava_Report(errPtr, t);
    if (errPtr == 0) {
        SignalError(EE(), "xerox/ilu/IluSomeSystemException", t);
        return;
    }
    ILU_ERR_SWITCH(*errPtr) {
        ILU_SUCCESS_CASE return;

        /*.../ilu/runtime/kernel/iluerrs.h calls these public */
        E_SWITCH(unknown,"xerox/ilu/IluSomeSystemException")
        E_SWITCH(bad_param,"xerox/ilu/IluBadParameterException")
        UNIMPL_SWITCH(no_memory,CORBAPREFIX "NO_MEMORY")
        UNIMPL_SWITCH(imp_limit,CORBAPREFIX "IMP_LIMIT")
        UNIMPL_SWITCH(comm_failure,CORBAPREFIX "COMM_FAILURE")
        UNIMPL_SWITCH(inv_objref,CORBAPREFIX "INV_OBJREF")
        UNIMPL_SWITCH(no_permission,CORBAPREFIX "NO_PERMISSION")
        E_SWITCH(internal,"xerox/ilu/IluInternalException")
        E_SWITCH(marshal,CORBAPREFIX "MARSHAL")
        UNIMPL_SWITCH(initialize,CORBAPREFIX "INITIALIZE")
        UNIMPL_SWITCH(no_implement,CORBAPREFIX "NO_IMPLEMENT")
        UNIMPL_SWITCH(bad_typecode,CORBAPREFIX "BAD_TYPECODE")
        UNIMPL_SWITCH(bad_operation,CORBAPREFIX "BAD_OPERATION")
        UNIMPL_SWITCH(no_resources,CORBAPREFIX "NO_RESOURCES")
        UNIMPL_SWITCH(no_response,CORBAPREFIX "NO_RESPONSE")
        UNIMPL_SWITCH(persist_store,CORBAPREFIX "PERSIST_STORE")
        UNIMPL_SWITCH(bad_inv_order,CORBAPREFIX "BAD_INV_ORDER")
        UNIMPL_SWITCH(transient,CORBAPREFIX "TRANSIENT")
        UNIMPL_SWITCH(free_mem,CORBAPREFIX "FREE_MEM")
        UNIMPL_SWITCH(inv_ident,CORBAPREFIX "INV_IDENT")
        UNIMPL_SWITCH(inv_flag,CORBAPREFIX "INV_FLAG")
        UNIMPL_SWITCH(intf_repos,CORBAPREFIX "INTF_REPOS")
        UNIMPL_SWITCH(bad_context,CORBAPREFIX "BAD_CONTEXT")
        UNIMPL_SWITCH(obj_adapter,CORBAPREFIX "OBJ_ADAPTER")
        UNIMPL_SWITCH(data_conversion,CORBAPREFIX "DATA_CONVERSION")
        UNIMPL_SWITCH(bad_locks,CORBAPREFIX "BAD_LOCKS")
        UNIMPL_SWITCH(broken_locks,CORBAPREFIX "BROKEN_LOCKS")
        UNIMPL_SWITCH(interrupted,CORBAPREFIX "INTERRUPTED")
        UNIMPL_SWITCH(gss_security,CORBAPREFIX "GSS_SECURITY")
        
        /*.../ilu/runtime/kernel/iluerrs.h calls these unresolved */
        E_SWITCH(KernelBroken,"xerox/ilu/IluInternalException")
        E_SWITCH(BadDataStructure,"xerox/ilu/IluInternalException")
        E_SWITCH(MallocFailure,"xerox/ilu/IluInternalException")
        E_SWITCH(MalformedKernelObject,"xerox/ilu/IluInternalException")
        E_SWITCH(MaxCountExceeded,"xerox/ilu/IluInternalException")
        E_SWITCH(ProtocolAlreadyRegistered,"xerox/ilu/IluSomeSystemException")
        E_SWITCH(TransportAlreadyRegistered,"xerox/ilu/IluSomeSystemException")
        E_SWITCH(WrongLocks,"xerox/ilu/IluInternalException")
        E_SWITCH(BadProtocolInfo,"xerox/ilu/IluSomeSystemException")
        E_SWITCH(BadArguments,"xerox/ilu/IluSomeSystemException")
        E_SWITCH(GcRegFailed,"xerox/ilu/IluSomeSystemException")
        E_SWITCH(NoConnection,"xerox/ilu/IluSomeSystemException")
        E_SWITCH(ClosedServer,"xerox/ilu/IluSomeSystemException")
        E_SWITCH(ObjectNotFoundInServer,"xerox/ilu/IluSomeSystemException")
        E_SWITCH(NoObjectForSBH,"xerox/ilu/IluSomeSystemException")
        E_SWITCH(CantCondition,"xerox/ilu/IluSomeSystemException")
        
        ILU_ERR_ELSE {
            SignalError(EE(), "xerox/ilu/IluSomeSystemException", t);
            return;
        }
    } ILU_ERR_ENDSWITCH;
}


EXPORTLIMITED void
_ilujava_NullPointerException(char* t)
/* Causes a NullPointerException to be reported when returned into java.
 */
{
    if (t==0) t = "unspecified";
    _ilujava_Report(0, "NullPointerException");
    SignalError(EE(), "java/lang/NullPointerException", t);
}
 
 
EXPORTLIMITED void
_ilujava_SignalInconsistency(char* t)
/* Causes a serious internal error to be reported when returned into java.
 */
{
    if (t==0) t = "unspecified";
    ilu_DebugPrintf("$ IluCall inconsistency\n");
    SignalError(EE(), "xerox/ilu/IluInconsistentCallException", t);
}


/* ================  ================ */


EXPORTJAVA JIluOInt
xerox_ilu_IluRT0_oiFromSBH(
	JIluRT0 unused,
	JString jSBH,
	JIluClassRep jjClassRep
	)
{
    JIluOInt jjoi; 
    ilu_Object cIluObject;
    ilu_Error err = ILU_INIT_NO_ERR;
    ilu_Class cIluClass;
    char * cSBH;
    ilu_ConsiderSbhResult result;
    
    if (jSBH == 0) {
        return 0;
    }
    
    if (jjClassRep == 0) {
        cIluClass = ilu_rootClass;
    } else {
        cIluClass = Access(jjClassRep, yIluClass, ilu_Class);
    }
        
    cSBH = makeCString(jSBH); /*<javaString.h> subject to GC*/
    
    result = ilu_ConsiderSBH(cSBH, &err);
    if (result == ilucsr_err) {
        _ilujava_IluErrorToException(&err, "ilu_ConsiderSBH");
        return 0;
    }
    
    cIluObject = ilu_ObjectOfSBH(cSBH, cIluClass, &err);
    if (ILU_ERRNOK(err)) {
        _ilujava_IluErrorToException(&err, "ilu_ObjectOfSBH");
        return 0;
    }
    if (cIluObject == 0) {
        return 0;
    }
    jjoi = (JIluOInt) 
        ilu_GetLanguageSpecificObject(cIluObject, JAVALANGIDX);
    /* true or surrogate at this point */
    if (jjoi == 0) {
        /* but we do not create true objects here */
        jjoi = (JIluOInt) 
            _ilujava_creatSurrOIFromRegistry(
        	cIluObject, 
        	(ilu_refany) jjClassRep
        	);
    } else {
        if (unhand(jjoi)->retained) {unhand(jjoi)->ghost = 1;}
    }
    ilu_ExitServer(
    	ilu_ServerOfObject(cIluObject),
    	cIluClass
    	);
    _ilujava_notUsed1 = cSBH;
    return jjoi; 
}


EXPORTJAVA JIluOInt
xerox_ilu_IluRT0_oiFromLookup(
	JIluRT0 unused,
	JString jsid,
	JString jjih,
	JIluClassRep jjClassRep
	)
{
    JIluOInt jjoi; 
    char * csid;
    char * cih;
    ilu_Object cIluObject = ILU_NIL;  
    ilu_Class cIluClass;
    ilu_boolean change;
    if (jsid == 0 OR jjih == 0) {
        return 0;
    }
    csid = makeCString(jsid); /* <javaString.h> subject to GC */
    cih = makeCString(jjih); /* <javaString.h> subject to GC */
    if (jjClassRep == 0) {
        cIluClass = ilu_rootClass;
    } else {
        cIluClass = Access(jjClassRep, yIluClass, ilu_Class);
    }
    change = ilu_ReLookupObject(csid, cih, cIluClass, &cIluObject);
    if (cIluObject == 0) {
        return 0;
    }
    jjoi = (JIluOInt) ilu_GetLanguageSpecificObject(cIluObject, JAVALANGIDX);
    if (jjoi == ILU_NIL) {
        jjoi = (JIluOInt) _ilujava_creatSurrOIFromRegistry(
        	cIluObject, 
        	(ilu_refany) jjClassRep
        	);
    } else {
        if (unhand(jjoi)->retained) {unhand(jjoi)->ghost = 1;}
    }
    ilu_ExitServer(
    	ilu_ServerOfObject(cIluObject),
    	cIluClass
    	);
    _ilujava_notUsed1 = csid;
    _ilujava_notUsed2 = cih;
    return jjoi;
}


EXPORTJAVA JString
xerox_ilu_IluRT0_inventID(JIluRT0 unused)
{
    JString jjstring;
    ilu_string cstring = ilu_InventID();
    jjstring = /*javaString.h*/ makeJavaString(cstring, strlen(cstring));
    ilu_free(cstring);
    return jjstring;
}

IMPORTVAR char * _ilujava_notUsed1;

EXPORTJAVA void
xerox_ilu_IluRT0_registerTrue(
	JIluRT0 unused,
	JIluOInt jjoi,
	JString jjih,
	JIluServer jjIluServer
	)
/* NOTE: 
 * similar code in (IluJava_IluServer.c)_ilujava_ObjectTableCreateObj
 */
{
    ilu_Object cObject;
    ilu_string cih = 0;
    ilu_Server cServer;
    ilu_Class cClass;
    JIluClassRep jjClass;
    if ((jjoi == 0) | (jjIluServer == 0) | (jjih == 0)) {
        BUGPRINT("$ ** IluRT0_registerTrue ERROR BAD ARGUMENTS\n");
        return;
    }
    jjClass = unhand(jjoi)->jjClassRep;
    if (jjClass == 0) {
        BUGPRINT("$ ** IluRT0_registerTrue ERROR no ilu-class\n");
        return;
    }
    cClass = Access(jjClass, yIluClass, ilu_Class);
    if (cClass == 0) {
        BUGPRINT("$ ** IluRT0_registerTrue ERROR bad ilu-class\n");
        return;
    }
    cServer = _ilujava_EnterServer2(jjIluServer, cClass);
    if (cServer==0) {
        BUGPRINT("$ ** IluRT0_registerTrue ERROR server destroyed\n");
        return;
    }
    Access(jjoi, yServer, ilu_Server) = cServer;
    
    cih = allocCString(jjih); /*javaString.h using malloc*/ 
    cObject = ilu_FindOrCreateTrueObject(cih, cServer, cClass, jjoi);
    java_free(cih);
    if (cObject) {
        ilu_refany jalt;
        if (_ilujava_gcFlag > 1) {
            cObject->ob_timeout = 120; /*for debugging: gc after two minutes*/
        }
        jalt = ilu_GetLanguageSpecificObject(cObject, JAVALANGIDX);
        if (jalt != jjoi) {
            /* This can happen: 
             * jjoi has been created outside of the server lock.... 
             */
            if (_ilujava_objectsFlag > 1) {
                ilu_DebugPrintf("$ registerTrue: object creation conflict\n");   
            }
            cObject = 0;
        }
    }
    if (_ilujava_objectsFlag > 1) {
        ilu_DebugPrintf("$ IluRT0_registerTrue: found %x \n", cObject);   
    }
    Access(jjoi, yIluKernelObject, ilu_Object) = cObject;
    ilu_ExitServer(cServer, cClass);
    _ilujava_notUsed1 = cih;
}

/* end IluJava_IluRT0.c */
